/**************************************************************************************************
	$Header: /pub/cvsroot/mydns/lib/error.c,v 1.19 2002/08/04 03:37:31 bboy Exp $
	error.c: Error reporting and message output functions.

	Copyright (C) 2002  Don Moore <bboy@bboy.net>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at Your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**************************************************************************************************/

#include "mydnsutil.h"

/* The name of the currently running program */
char	*progname = PACKAGE;

/* Should ERR_VERBOSE cause output? */
int	err_verbose = 0;

/* Should ERR_DEBUG cause output? */
int	err_debug = 0;

/* Should all error.c functions output in a standard way (no pid, no source file, etc) */
int	err_plain = 0;

/* Should error functions output to stderr only? */
int	err_stderr = 0;


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	ERROR_INIT
	Initialize error reporting and output functions.
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void
error_init(argv0, facility)
	const char *argv0;	/* argv[0] from calling program */
	int facility;			/* Logging facility (i.e. LOG_DAEMON, etc) */
{
	int option;

	/* Save program name */
	if (argv0)
	{
		char *c;
		if ((c = strrchr(argv0, '/')))
			progname = xstrdup(c+1);
		else
			progname = xstrdup((char *)argv0);
	}

	/* Open syslog */
	if (!err_stderr)
	{
		option = LOG_CONS | LOG_NDELAY | LOG_PID;
		if (isatty(STDERR_FILENO))
			option |= LOG_PERROR;
		openlog(progname, option, facility);
	}
}
/*--- error_init() ------------------------------------------------------------------------------*/


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	_ERROR_OUT
	This function handles all program output after command line handling.
	Always returns -1.
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static void
_error_out(
	int priority,				/* Priority (i.e. LOG_ERR, LOG_WARNING, LOG_NOTICE, LOG_INFO, LOG_DEBUG) */
	int errappend,				/* Append strerror(errno)? */
	int isfatal,				/* Is this error fatal? */
	const char *msg 
)
{
	if (err_plain)
	{
		if (err_stderr)
		{
			if (errappend)
				fprintf(stderr, "%s: %s: %s\n", progname, msg, strerror(errno));
			else
				fprintf(stderr, "%s: %s\n", progname, msg);
		}
		else
			syslog(priority, errappend ? "%s: %m" : "%s", msg);
	}
	else if (priority == LOG_DEBUG)
	{
		if (err_stderr)
		{
			if (errappend)
				fprintf(stderr, "%s: %s: %s\n", progname, msg, strerror(errno));
			else
				fprintf(stderr, "%s: %s\n", progname, msg);
		}
		else
			syslog(priority, errappend ? "%s: %m" : "%s", msg);
	}
	else
	{
		if (err_stderr)
		{
			if (errappend)
				fprintf(stderr, "%s: %s: %s\n", progname, msg, strerror(errno));
			else
				fprintf(stderr, "%s: %s\n", progname, msg);
		}
		else
			syslog(priority, errappend ? "%s: %m" : "%s", msg);
	}

	/* Abort if the error should be fatal */
	if (isfatal)
		exit(EXIT_FAILURE);
}
/*--- _error_out() ------------------------------------------------------------------------------*/


/**************************************************************************************************
   _ERROR_ASSERT_FAIL
   This is called by the macro "Assert" if an assertion fails.
**************************************************************************************************/
void
_error_assert_fail(assertion)
	const char *assertion;
{
	char msg[BUFSIZ];

	snprintf(msg, sizeof(msg), "assertion failed: %s", assertion);

   _error_out(LOG_ERR, 0, 1, msg);
}
/*--- _error_assert_fail() ----------------------------------------------------------------------*/


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	DEBUG
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void
Debug(const char *fmt, ...)
{
	char msg[BUFSIZ];
	va_list ap;

	if (!err_debug)
		return;

	/* Construct output string */
	va_start(ap, fmt);
	vsnprintf(msg, sizeof(msg), fmt, ap);
	va_end(ap);

	_error_out(LOG_DEBUG, 0, 0, msg);
}
/*--- Debug() -----------------------------------------------------------------------------------*/


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	VERBOSE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void
Verbose(const char *fmt, ...)
{
	char msg[BUFSIZ];
	va_list ap;

	if (!err_verbose)
		return;

	/* Construct output string */
	va_start(ap, fmt);
	vsnprintf(msg, sizeof(msg), fmt, ap);
	va_end(ap);

	_error_out(LOG_NOTICE, 0, 0, msg);
}
/*--- Verbose() ---------------------------------------------------------------------------------*/


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	NOTICE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void
Notice(const char *fmt, ...)
{
	char msg[BUFSIZ];
	va_list ap;

	/* Construct output string */
	va_start(ap, fmt);
	vsnprintf(msg, sizeof(msg), fmt, ap);
	va_end(ap);

	_error_out(LOG_NOTICE, 0, 0, msg);
}
/*--- Notice() ----------------------------------------------------------------------------------*/


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	WARN
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int
Warn(const char *fmt, ...)
{
	char msg[BUFSIZ];
	va_list ap;

	/* Construct output string */
	va_start(ap, fmt);
	vsnprintf(msg, sizeof(msg), fmt, ap);
	va_end(ap);

	_error_out(LOG_WARNING, 1, 0, msg);
	return (-1);
}
/*--- Warn() ------------------------------------------------------------------------------------*/


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	WARNX
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int
Warnx(const char *fmt, ...)
{
	char msg[BUFSIZ];
	va_list ap;

	/* Construct output string */
	va_start(ap, fmt);
	vsnprintf(msg, sizeof(msg), fmt, ap);
	va_end(ap);

	_error_out(LOG_WARNING, 0, 0, msg);
	return (-1);
}
/*--- Warnx() -----------------------------------------------------------------------------------*/


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	ERR
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void
Err(const char *fmt, ...)
{
	char msg[BUFSIZ];
	va_list ap;

	/* Construct output string */
	va_start(ap, fmt);
	vsnprintf(msg, sizeof(msg), fmt, ap);
	va_end(ap);

	_error_out(LOG_ERR, 1, 1, msg);
}
/*--- Err() -------------------------------------------------------------------------------------*/


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	ERRX
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void
Errx(const char *fmt, ...)
{
	char msg[BUFSIZ];
	va_list ap;

	/* Construct output string */
	va_start(ap, fmt);
	vsnprintf(msg, sizeof(msg), fmt, ap);
	va_end(ap);

	_error_out(LOG_ERR, 0, 1, msg);
}
/*--- Errx() ------------------------------------------------------------------------------------*/


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	WARNSQL
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int
WarnSQL(MYSQL *mysql, const char *fmt, ...)
{
	char msg[BUFSIZ], out[BUFSIZ * 2];
	va_list ap;

	/* Construct output string */
	va_start(ap, fmt);
	vsnprintf(msg, sizeof(msg), fmt, ap);
	va_end(ap);

	snprintf(out, sizeof(out), "%s: %s (errno=%d)", msg, mysql_error(mysql), errno);
	_error_out(LOG_WARNING, 0, 0, out);

	return (-1);
}
/*--- WarnSQL() ---------------------------------------------------------------------------------*/


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	ERRSQL
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void
ErrSQL(MYSQL *mysql, const char *fmt, ...)
{
	char msg[BUFSIZ], out[BUFSIZ * 2];
	va_list ap;

	/* Construct output string */
	va_start(ap, fmt);
	vsnprintf(msg, sizeof(msg), fmt, ap);
	va_end(ap);

	snprintf(out, sizeof(out), "%s: %s (errno=%d)", msg, mysql_error(mysql), errno);
	_error_out(LOG_ERR, 0, 1, out);
}
/*--- ErrSQL() ----------------------------------------------------------------------------------*/

/* vi:set ts=3: */
