/**************************************************************************************************
	$Header: /pub/cvsroot/mydns/src/mydns/axfr.c,v 1.8 2002/09/20 19:43:40 bboy Exp $
	axfr.c: Routines to do zone transfers.

	Copyright (C) 2002  Don Moore <bboy@bboy.net>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at Your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**************************************************************************************************/

#include "named.h"

#define	AXFR_TIME_LIMIT		3600		/* AXFR may not take more than this long, overall */

static size_t total_records, total_octets;


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	AXFR_ERROR
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static void axfr_error(TASK *, const char *, ...) __attribute__ ((__noreturn__));
static void
axfr_error(TASK *t, const char *fmt, ...)
{
	va_list ap;
	char msg[BUFSIZ];

	va_start(ap, fmt);
	vsnprintf(msg, sizeof(msg), fmt, ap);
	va_end(ap);

	if (t)
		Warnx("%s: %s", desctask(t), msg);
	else
		Warnx("%s", msg);

#if HAVE_SHUTDOWN
	shutdown(t->fd, 2);
#endif
	close(t->fd);
	_exit(EXIT_FAILURE);
}
/*--- axfr_error() ------------------------------------------------------------------------------*/


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	AXFR_TIMEOUT
	Hard timeout called by SIGALRM after one hour.
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static void
axfr_timeout(int dummy)
{
	axfr_error(NULL, _("AXFR timed out"));
}
/*--- axfr_timeout() ----------------------------------------------------------------------------*/


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	AXFR_WRITE_WAIT
	Wait for the client to become ready to read.  Times out after `task_timeout' seconds.
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static void
axfr_write_wait(t)
	TASK *t;
{
	fd_set wfd;
	struct timeval tv;
	int rv;

	FD_ZERO(&wfd);
	FD_SET(t->fd, &wfd);
	tv.tv_sec = task_timeout;
	tv.tv_usec = 0;
	if ((rv = select(t->fd+1, NULL, &wfd, NULL, &tv)) < 0)
		axfr_error(t, "%s: %s", _("select"), strerror(errno));
	if (rv != 1 || !FD_ISSET(t->fd, &wfd))
		axfr_error(t, _("write timeout"));
}
/*--- axfr_write_wait() -------------------------------------------------------------------------*/


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	AXFR_WRITE
	Writes the specified buffer, obeying task_timeout (via axfr_write_wait).
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static void
axfr_write(t, buf, size)
	TASK *t;
	u_char *buf;
	size_t size;
{
	int rv, offset = 0;

	do
	{
		axfr_write_wait(t);
		if ((rv = write(t->fd, buf+offset, size-offset)) < 0)
			axfr_error(t, "write: %s", strerror(errno));
		if (!rv)
			axfr_error(t, _("client closed connection"));
		offset += rv;
	} while (offset < size);
}
/*--- axfr_write() ------------------------------------------------------------------------------*/


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	AXFR_REPLY
	Sends one reply to the client.
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static void
axfr_reply(t)
	TASK *t;
{
	u_char len[2], *l = len;

	Assert(t);
	build_reply(t, 0);
	DNS_PUT16(l, t->replylen);
	axfr_write(t, len, SIZE16);
	axfr_write(t, t->reply, t->replylen);
	total_octets += SIZE16 + t->replylen;
	total_records++;

	/* Reset the pertinent parts of the task reply data */
	rrlist_free(&t->an);
	rrlist_free(&t->ns);
	rrlist_free(&t->ar);

	if (t->reply) free(t->reply);
	t->reply = NULL;
	t->replylen = 0;

	name_forget(t);

	if (t->rdata) free(t->rdata);
	t->rdata = NULL;
	t->rdlen = 0;

	/* Nuke question data */
	t->qdcount = 0;
	t->qdlen = 0;
}
/*--- axfr_reply() ------------------------------------------------------------------------------*/


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	CHECK_XFER
	If the "xfer" column exists in the soa table, it should contain a list of wildcards separated
	by commas.  In order for this zone transfer to continue, one of the wildcards must match
	the client's IP address.
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static void
check_xfer(t, soa)
	TASK *t;
	MYDNS_SOA *soa;
{
	MYSQL_RES *res = NULL;
	MYSQL_ROW row;
	char *ip = inet_ntoa(t->addr.sin_addr);
	char query[512];
	int  ok = 0;

	if (!mydns_soa_use_xfer)
		return;

	snprintf(query, sizeof(query), "SELECT xfer FROM %s WHERE id=%u%s",
		mydns_soa_table_name, soa->id, mydns_rr_use_active ? " AND active=1" : "");

	if (mysql_real_query(mysql, query, strlen(query)) || !(res = mysql_use_result(mysql)))
		ErrSQL(mysql, "%s: %s", desctask(t), _("error loading zone transfer access rules"));

	if ((row = mysql_fetch_row(res)))
	{
		char *wild, *r;

		for (r = row[0]; (wild = strsep(&r, ",")); )
			if (wildcard_match(wild, ip))
				ok = 1;
	}
	mysql_free_result(res);

	if (!ok)
	{
		dnserror(t, DNS_R_REFUSED, ERR_NO_AXFR);
		axfr_reply(t);
		axfr_error(t, _("access denied"));
	}
}
/*--- check_xfer() ------------------------------------------------------------------------------*/


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	AXFR_ZONE
	Transfer a non-PTR zone.
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static void
axfr_zone(t)
	TASK *t;
{
	MYDNS_SOA *soa = NULL;
	MYSQL_RES *res = NULL;
	MYSQL_ROW row;
	char query[256];

	/* Load the SOA for the requested zone */
	if (mydns_soa_load(mysql, &soa, t->qname) < 0)
		ErrSQL(mysql, "%s: %s", desctask(t), _("error loading zone"));
	if (!soa)
	{
		dnserror(t, DNS_R_REFUSED, ERR_ZONE_NOT_FOUND);
		axfr_reply(t);
		axfr_error(t, _("unknown zone"));
	}
	check_xfer(t, soa);											/* Check optional access column */
	reply_init(t);

	/* Send opening SOA record */
	rrlist_add(t, ANSWER, DNS_RRTYPE_SOA, (void *)soa, soa->origin);
	axfr_reply(t);

	/* Get all resource records for zone */
	snprintf(query, sizeof(query),
				"SELECT "MYDNS_RR_FIELDS" FROM %s WHERE zone=%u%s",
				mydns_rr_table_name, soa->id, mydns_rr_use_active ? " AND active=1" : "");
	if (mysql_real_query(mysql, query, strlen(query)) || !(res = mysql_use_result(mysql)))
		ErrSQL(mysql, "%s: %s", desctask(t), _("error loading resource records"));
	while ((row = mysql_fetch_row(res)))
	{
		MYDNS_RR *rr;
		int len;

		if (!(rr = mydns_parse_rr(row)))
			continue;
		len = strlen(rr->name);
		if (rr->name[len-1] != '.')
		{
			if (*rr->name)
				strcat(rr->name, ".");
			strncat(rr->name, soa->origin, sizeof(rr->name) - len - 1);
		}
		/* Append origin to data for certain types */
		switch (rr->type)
		{
			case DNS_T_CNAME:
			case DNS_T_NS:
			case DNS_T_MX:
				len = strlen(rr->data);
				if (rr->data[len-1] != '.')
					strncat(rr->data, soa->origin, sizeof(rr->data) - len - 1);
				break;
			default:
				break;
		}
		rrlist_add(t, ANSWER, DNS_RRTYPE_RR, (void *)rr, rr->name);
		axfr_reply(t);
		mydns_rr_free(rr);
	}
	mysql_free_result(res);

	/* Send closing SOA record */
	rrlist_add(t, ANSWER, DNS_RRTYPE_SOA, (void *)soa, soa->origin);
	axfr_reply(t);

	mydns_soa_free(soa);
}
/*--- axfr_zone() -------------------------------------------------------------------------------*/


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	AXFR_PTR_ZONE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static void
axfr_ptr_zone(t, label, origin)
	TASK *t;
	u_char *label;
	u_char *origin;
{
	u_char ip[4];
	MYDNS_SOA *soa;
	MYSQL_RES *res = NULL;
	MYSQL_ROW row;
	struct tm *tm;
	char query[256], serial[40];

	/* Split in-addr.arpa format address into quads */
	switch (mydns_extract_arpa(origin, ip))
	{
		case 3:
			snprintf(query, sizeof(query),
				"SELECT "MYDNS_PTR_FIELDS" FROM %s WHERE "
				"ip >= INET_ATON('%d.%d.%d.0') AND ip <= INET_ATON('%d.%d.%d.255')%s",
				mydns_ptr_table_name, ip[2], ip[1], ip[0], ip[2], ip[1], ip[0],
				mydns_ptr_use_active ? " AND active=1" : "");
			break;

		case 2:
			snprintf(query, sizeof(query),
				"SELECT "MYDNS_PTR_FIELDS" FROM %s WHERE "
				"ip >= INET_ATON('%d.%d.0.0') AND ip <= INET_ATON('%d.%d.255.255')%s",
				mydns_ptr_table_name, ip[1], ip[0], ip[1], ip[0],
				mydns_ptr_use_active ? " AND active=1" : "");
			break;

		case 1:
			snprintf(query, sizeof(query),
				"SELECT "MYDNS_PTR_FIELDS" FROM %s WHERE "
				"ip >= INET_ATON('%d.0.0.0') AND ip <= INET_ATON('%d.255.255.255')%s",
				mydns_ptr_table_name, ip[0], ip[0],
				mydns_ptr_use_active ? " AND active=1" : "");
			break;

		default:
			dnserror(t, DNS_R_REFUSED, ERR_NAME_FORMAT);
			axfr_reply(t);
			axfr_error(t, _("Invalid zone format"));
	}

	/* Get all ptr records for zone */
	if (mysql_real_query(mysql, query, strlen(query)) || !(res = mysql_store_result(mysql)))
		ErrSQL(mysql, "%s: %s", desctask(t), _("error loading ptr records"));

	/* If no results were found, refuse request */
	if (!mysql_num_rows(res))
	{
		mysql_free_result(res);
		dnserror(t, DNS_R_REFUSED, ERR_ZONE_NOT_FOUND);
		axfr_reply(t);
		axfr_error(t, _("zone not found"));
	}

	/* Send opening SOA record */
	if (mydns_soa_load(mysql, &soa, origin) != 0)
		ErrSQL(mysql, "%s: %s", label, _("error loading zone"));
 	if (!soa && mydns_soa_make(mysql, &soa, origin, hostname) != 0)
		ErrSQL(mysql, "%s: %s", label, _("error loading zone"));
	if (!soa)
	{
		dnserror(t, DNS_R_REFUSED, ERR_ZONE_NOT_FOUND);
		axfr_reply(t);
		axfr_error(t, _("unknown zone"));
	}

	name_forget(t);
#if DEBUG
	Debug("%s: %s", desctask(t), _("starting zone transfer"));
#endif
	rrlist_add(t, ANSWER, DNS_RRTYPE_SOA, (void *)soa, soa->origin);
	axfr_reply(t);

	/* Now send the PTR records */
	while ((row = mysql_fetch_row(res)))
	{
		MYDNS_PTR *ptr;
		u_char q[4], fqdn[DNS_MAXNAMELEN+1];

		if (!(ptr = mydns_parse_ptr(row)))
			continue;

		memcpy(&q, &ptr->ip, sizeof(q));
		snprintf(fqdn, sizeof(fqdn), "%d.%d.%d.%d" PTR_SUFFIX, q[0], q[1], q[2], q[3]);
		rrlist_add(t, ANSWER, DNS_RRTYPE_PTR, (void *)ptr, fqdn);
		axfr_reply(t);
		mydns_ptr_free(ptr);
	}
	mysql_free_result(res);

	/* Send closing SOA record */
	rrlist_add(t, ANSWER, DNS_RRTYPE_SOA, (void *)soa, soa->origin);
	axfr_reply(t);
}
/*--- axfr_ptr_zone() ---------------------------------------------------------------------------*/


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	AXFR
	DNS-based zone transfer.  Send all resource records for in QNAME's zone to the client.
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void
axfr(t)
	TASK *t;
{
	struct timeval start, finish;
	int len;

	Assert(t);

	signal(SIGALRM, axfr_timeout);
	alarm(AXFR_TIME_LIMIT);
	db_connect();
	gettimeofday(&start, NULL);
	total_records = total_octets = 0;
	t->no_markers = 1;

	/* If the zone requested ends in PTR_SUFFIX, do ptr axfr, otherwise normal zone */
	len = strlen(t->qname);
	if (len > PTR_SUFFIX_LEN && !strcmp(t->qname + (len-PTR_SUFFIX_LEN), PTR_SUFFIX))
	{
		u_char label[DNS_MAXNAMELEN+1];

		memcpy(label, t->qname, len-PTR_SUFFIX_LEN);
		label[len-PTR_SUFFIX_LEN] = '\0';
		axfr_ptr_zone(t, label, t->qname);
	}
	else
		axfr_zone(t);

	gettimeofday(&finish, NULL);
#if DEBUG
	Debug(_("%s: zone transfer complete in %.3fs (%u records, %u octets)"), desctask(t),
		((finish.tv_sec + finish.tv_usec / 1000000.0) - (start.tv_sec + start.tv_usec / 1000000.0)),
		total_records, total_octets);
#endif

#if HAVE_SHUTDOWN
	shutdown(t->fd, 2);
#endif
	close(t->fd);
	_exit(EXIT_SUCCESS);
}
/*--- axfr() ------------------------------------------------------------------------------------*/

/* vi:set ts=3: */
/* NEED_PO */
