/**************************************************************************************************
	$Header: /pub/cvsroot/mydns/src/mydns/tcp.c,v 1.9 2002/09/17 13:24:33 bboy Exp $
	tcp.c: Routines for TCP I/O.

	Copyright (C) 2002  Don Moore <bboy@bboy.net>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at Your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**************************************************************************************************/

#include "named.h"


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	ACCEPT_TCP_QUERY
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int
accept_tcp_query(fd)
	int fd;
{
	struct sockaddr_in addr;
	int rmt_fd, addrlen = sizeof(addr);
	TASK *t;

	if ((rmt_fd = accept(fd, (struct sockaddr *)&addr, &addrlen)) < 0)
		return Warn("accept");
	fcntl(rmt_fd, F_SETFL, fcntl(rmt_fd, F_GETFL, 0) | O_NONBLOCK);

	t = task_init(NEED_READ, rmt_fd, SOCK_STREAM, &addr);
	enqueue(Tasks, t);

#if DEBUG
	Debug(
#if PRINTF_NEEDS_QUAD
		"[%qu] %s: %s",
#else
		"[%llu] %s: %s",
#endif
		t->internal_id, inet_ntoa(t->addr.sin_addr), _("TCP connection accepted"));
#endif

	return 0;
}
/*--- accept_tcp_query() ------------------------------------------------------------------------*/


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	READ_TCP_LENGTH
	The first two octets of a TCP question are the length.  Read them.
	Returns 0 on success, -1 on failure.
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static int
read_tcp_length(t)
	TASK *t;
{
	int rv;
	u_char len[2];

	if ((rv = recv(t->fd, len, 2, 0)) != 2)
	{
		if (rv < 0)
			return Warn("%s: recv (length)", inet_ntoa(t->addr.sin_addr));
		return Warnx("%s: %s", inet_ntoa(t->addr.sin_addr), _("TCP message length invalid"));
	}

	if ((t->len = ((len[0] << 8) | (len[1]))) < DNS_HEADERSIZE)
		return Warnx("%s: %s", inet_ntoa(t->addr.sin_addr), _("TCP message too short"));
	if (t->len > DNS_MAXPACKETLEN_TCP)
		return Warnx("%s: %s", inet_ntoa(t->addr.sin_addr), _("TCP message too long"));

	if (!(t->query = calloc(1, t->len + 1)))
		Err("calloc");
	t->offset = 0;
	return (0);
}
/*--- read_tcp_length() -------------------------------------------------------------------------*/


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	READ_TCP_QUERY
	Returns 0 on success, -1 on failure.
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int
read_tcp_query(t)
	TASK *t;
{
	u_char *src, *qdtop;
	int rv;
	u_char qname[DNS_MAXNAMELEN+1];
	u_int16_t ancount, nscount, arcount;

	/* Read packet length if we haven't already */
	if (!t->len)
		return read_tcp_length(t);

	/* Read whatever data is ready */
	if ((rv = recv(t->fd, t->query + t->offset, t->len - t->offset, 0)) < 0)
		return Warn("%s: %s", inet_ntoa(t->addr.sin_addr), _("recv"));
	if (!rv)
		return Warnx("%s: %s", inet_ntoa(t->addr.sin_addr), _("TCP client closed connection"));
#if DEBUG
	Debug(
#if PRINTF_NEEDS_QUAD
		"[%qu] %s: 2+%d %s",
#else
		"[%llu] %s: 2+%d %s",
#endif
		t->internal_id, inet_ntoa(t->addr.sin_addr), rv, _("octets in"));
#endif
	t->offset += rv;
	if (t->offset > t->len)
		return Warnx("%s: %s", inet_ntoa(t->addr.sin_addr), _("TCP message data too long"));
	if (t->offset < t->len)
		return 0;													/* Not finished reading */

	/* Reset offset for writing reply */
	t->offset = 0;

	/* Parse query header data */
	src = t->query;
	DNS_GET16(t->id, src);
	memcpy(&t->hdr, src, SIZE16); src += SIZE16;
	DNS_GET16(t->qdcount, src);
	DNS_GET16(ancount, src);
	DNS_GET16(nscount, src);
	DNS_GET16(arcount, src);

#if DEBUG
	Debug("%s: id=%u qr=%u opcode=%s aa=%u tc=%u rd=%u ra=%u z=%u rcode=%u", desctask(t),
			t->id, t->hdr.qr, mydns_opcode_str(t->hdr.opcode),
			t->hdr.aa, t->hdr.tc, t->hdr.rd, t->hdr.ra, t->hdr.z, t->hdr.rcode);
	Debug("%s: qd=%u an=%u ns=%u ar=%u", desctask(t),
			t->qdcount, ancount, nscount, arcount);
#endif

	/* Check validity of header fields */
	if (!ntohs(t->qdcount))
		return Warnx("%s: %s", inet_ntoa(t->addr.sin_addr), _("message does not contain any questions"));
	if (t->hdr.tc)
		return Warnx("%s: %s", inet_ntoa(t->addr.sin_addr), _("truncated request"));
	if (ntohs(t->hdr.opcode) != DNS_OPCODE_QUERY)
		return Warnx("%s: %s", inet_ntoa(t->addr.sin_addr), _("not a standard query"));

	/* Fill in question data */
	t->qdlen = t->len - DNS_HEADERSIZE;
	if (t->qdlen <= 0)
		return Warnx("%s: %s", inet_ntoa(t->addr.sin_addr), _("invalid question data length"));
	if (!(t->qd = (u_char *)malloc(t->qdlen)))
		Err("malloc");
	memcpy(t->qd, src, t->qdlen);
	qdtop = src;

	/* Get QNAME, QTYPE, and QCLASS */
	if (!(src = name_unencode(t->qd, t->qdlen, src, qname, sizeof(qname))))
		return Warnx("%s: %s", _("error getting name from question"), qname);
	strncpy(t->qname, qname, sizeof(t->qname)-1);
	DNS_GET16(t->qtype, src);
	DNS_GET16(t->qclass, src);
	t->qdlen = src - qdtop;

	if (!axfr_enabled && t->qtype == DNS_T_AXFR)
		return Warnx("%s: %s", inet_ntoa(t->addr.sin_addr), _("AXFR refused"));

	if (!tcp_enabled && t->qtype != DNS_T_AXFR)
		return Warnx("%s: %s", inet_ntoa(t->addr.sin_addr), _("TCP query refused"));

	/* Check class (only IN or ANY are allowed) */
	if ((t->qclass != DNS_C_IN) && (t->qclass != DNS_C_ANY))
	{
		dnserror(t, DNS_R_NOTIMP, ERR_NO_CLASS);
		build_reply(t, 1);
		return 0;
	}

#if DEBUG
	Debug("%s: qtype=%s qclass=%s qname=[%s]", desctask(t),
			mydns_qtype_str(t->qtype), mydns_class_str(t->qclass), t->qname);
	Debug("%s: %u+2 TCP octets, id %u", desctask(t), t->len, t->id);
#endif

	/* If this is AXFR, fork to handle it so that other requests don't block */
	if (t->qtype == DNS_T_AXFR)
	{
		pid_t pid, parent;

		parent = getpid();
		if ((pid = fork()) < 0)
			return Warn("%s: fork", inet_ntoa(t->addr.sin_addr));
		if (!pid)
		{
			kill(parent, SIGCHLD);
			axfr(t);
		}
		pause();
		return -1;
	}

	t->status = NEED_ANSWER;

	return 0;
}
/*--- read_tcp_query() --------------------------------------------------------------------------*/


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	WRITE_TCP_LENGTH
	Writes the length octets for TCP reply.  Returns 0 on success, -1 on failure.
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static int
write_tcp_length(t)
	TASK *t;
{
	u_char len[2], *l;
	int rv;

#if DEBUG
	Debug("%s: %s", desctask(t), _("sending reply length"));
#endif

	l = len;
	DNS_PUT16(l, t->replylen);

	if ((rv = write(t->fd, len + t->offset, SIZE16 - t->offset)) < 0)
	{
		if (errno == EINTR)
			return 0;
		return Warn("%s: write (length)", inet_ntoa(t->addr.sin_addr));
	}
	if (!rv)
		return Warnx("%s: %s", inet_ntoa(t->addr.sin_addr), _("TCP client closed connection"));
	t->offset += rv;
	if (t->offset >= SIZE16)
	{
		t->len_written = 1;
		t->offset = 0;
	}
	return 0;
}
/*--- write_tcp_length() ------------------------------------------------------------------------*/


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	WRITE_TCP_REPLY
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int
write_tcp_reply(t)
	TASK *t;
{
	int rv;

	if (!t->len_written)
		return write_tcp_length(t);

#if DEBUG
	Debug("%s: %s", desctask(t), _("sending reply"));
#endif

	if ((rv = write(t->fd, t->reply + t->offset, t->replylen - t->offset)) < 0)
	{
		if (errno == EINTR)
			return 0;
		return Warn("%s: write", inet_ntoa(t->addr.sin_addr));
	}
	if (!rv)
		return Warnx("%s: %s", inet_ntoa(t->addr.sin_addr), _("TCP client closed connection"));
	t->offset += rv;
	if (t->offset >= t->replylen)
	{
		int rmt_fd = t->fd;
		struct sockaddr_in addr;
		TASK *new;

		/* Task complete; dequeue */
		/* For BIND8 AXFR, we leave the connection open until the remote host closes it - reset
			the connection here and get ready to answer another */
#if DEBUG
		Debug("%s: reply 2+%u octets", desctask(t), t->replylen);
#endif
		memcpy(&addr, &t->addr, sizeof(struct sockaddr_in));

		dequeue(Tasks, t, 0);
		enqueue(Tasks, task_init(NEED_READ, rmt_fd, SOCK_STREAM, &addr));
	}
	return (0);
}
/*--- write_tcp_reply() -------------------------------------------------------------------------*/

/* vi:set ts=3: */
/* NEED_PO */
